/// vector2f.h
///
/// (c) 2008-2020 by Carsten Busse <carsten.busse@googlemail.com>
///                  Bastian Spiegel <bs@tkscript.de> (additional coding)
///     - Distributed under terms of the Lesser GNU General Public License (LGPL).
///       See COPYING and <http://www.gnu.org/licenses/licenses.html#LGPL> for further information.
///
///

#ifndef __VECTOR2F_H_
#define __VECTOR2F_H_


YG("Math");

/* @class Vector2f,Object
This class represents a single precision 2 dimensional coordinate vector.
 */
YC class _Vector2f : public YAC_Object {
  public:
   sF32 floats[2];

  public:
   _Vector2f(void);
   ~_Vector2f();

   YAC_POOLED(_Vector2f,YAC_POOL_PRIORITY_HIGH);

   sUI   YAC_VCALL yacArrayGetNumElements    (void);
   sUI   YAC_VCALL yacArrayGetMaxElements    (void);
   sUI   YAC_VCALL yacArrayGetElementType    (void);
   sUI   YAC_VCALL yacArrayGetElementByteSize(void);
   void  YAC_VCALL yacArraySet               (YAC_ContextHandle, sUI, YAC_Value *);
   void  YAC_VCALL yacArrayGet               (YAC_ContextHandle, sUI, YAC_Value *);
   void *YAC_VCALL yacArrayGetPointer        (void);
   sSI   YAC_VCALL yacTensorRank             (void);
   void  YAC_VCALL yacSerialize              (YAC_Object *, sUI);
   sUI   YAC_VCALL yacDeserialize            (YAC_Object *, sUI);
   void  YAC_VCALL yacOperatorClamp          (YAC_Object *, YAC_Object *);
   void  YAC_VCALL yacOperatorWrap           (YAC_Object *, YAC_Object *);
   //supports:
   //YAC_OP_CEQ, YAC_OP_CNE, YAC_OP_ASSIGN, YAC_OP_ADD = _add,
   //YAC_OP_SUB = _sub, YAC_OP_MUL = _mul, YAC_OP_NEG, YAC_OP_INIT
   void  YAC_VCALL yacOperator               (sSI, YAC_Object *, YAC_Value *);
   void  YAC_VCALL yacOperatorI              (sSI, sSI, YAC_Value *);
   void  YAC_VCALL yacOperatorF32            (sSI, sF32, YAC_Value *);
   sUI   YAC_VCALL yacOperatorPriority       (void) { return YAC_OP_PRIO_FLOAT; }
   sBool YAC_VCALL yacToString               (YAC_String *) const;

  public:
   // Internal helper methods
   
   void assignFloats(sF32 *_fv, sUI _ne);
   void assignFloatArray(YAC_FloatArray *_fa);
   void assignGeneric(YAC_ContextHandle _ctx, YAC_Object *_o);

  public: 

   /* @method init,float x,float y
      Initialize vector
    */
   YM void init(sF32 _x, sF32 _y);

   //Element access methods
   /* @method setX,float a
      Set x element to a
      @arg a new value for x
    */
   YM void _setX(sF32);

   /* @method setY,float a
      Set y element to a
      @arg a new value for y
    */
   YM void _setY(sF32);

   /* @method getX:float
      Get value of x element
      @return value of x element
    */
   YM sF32 _getX(void) const;

   /* @method getY:float
      Get value of y element
      @return value of y element
    */
   YM sF32 _getY(void) const;

   //produce string representation
   /* @method getString:String
      Get string representation (x, y) of vector
      @return string representation of vector
    */
   YM void _getString(YAC_Value *) const;

   //arithmetic
   /* @method add,Vector2f v
      Add another vector v to this instance
      @arg v Vector to add
    */
   YM void _add_YAC_RSELF(YAC_Object *);
   YM void _add_YAC_RVAL(YAC_Object *, YAC_Value *);
   YM void _add_YAC_RARG(YAC_Object *, YAC_Object *) const;

   /* @method sub,Vector2f v
      Substract another vector v from this instance
      @arg v Vector to substract
    */
   YM void _sub_YAC_RSELF(YAC_Object *);
   YM void _sub_YAC_RVAL(YAC_Object *, YAC_Value *);
   YM void _sub_YAC_RARG(YAC_Object *, YAC_Object *) const;

   //scalar dot product
   /* @method dot,Vector2f v:float
      Calculate dot product from this instance and v
      @arg v Vector2f instance
      @return Scalar value
    */
   YM sF32 _dot(YAC_Object *);

   //tensor product - produces a 2x2 matrix from 2 Vector2f. Result is written to _retM2f
   /* @method tensor,Vector2f v:Matrix2f
      Create resulting matrix of tensor product from this instance and v
      @arg v Vector2f instance
      @return Matrix2f instance
    */
   YM void _tensor_YAC_RVAL(YAC_Object *, YAC_Value *);
   YM void _tensor_YAC_RARG(YAC_Object *_v2f, YAC_Object *_retM2f) const;

   /* @method addf,float v
      Add scalar value to all components
      @arg v scalar value
    */
   YM void _addf_YAC_RSELF(sF32);
   YM void _addf_YAC_RVAL(sF32, YAC_Value *);
   YM void _addf_YAC_RARG(sF32, YAC_Object *) const;

   /* @method mulf,float v
      Multiply this instance with scalar value v
      @arg v scalar value
    */
   YM void _mulf_YAC_RSELF(sF32);
   YM void _mulf_YAC_RVAL(sF32, YAC_Value *);
   YM void _mulf_YAC_RARG(sF32, YAC_Object *) const;

   //multiplies each element individual (not really mathematics, but sometimes useful)
   /* @method mul,Vector2f v
      Multiply the x element of this instance with the x element of v and do the same with the y element
      @arg v Vector2f instance
    */
   YM void _mul_YAC_RSELF(YAC_Object *);
   YM void _mul_YAC_RVAL(YAC_Object *, YAC_Value *);
   YM void _mul_YAC_RARG(YAC_Object *, YAC_Object *) const;

   /* @method getAbs:float
      Return length of vector
      @return length of vector
    */
   YM sF32 _getAbs(void) const;

   /* @method getAbsSqr:float
      Return squared length of vector
      @return length<sup>2</sup> of vector
    */
   YM sF32 _getAbsSqr(void) const;

   /* @method unit
      Scale vector to length 1 if possible
    */
   YM void _unit_YAC_RSELF(void); /* unit */
   YM void _unit_YAC_RVAL(YAC_Value *);
   YM void _unit_YAC_RARG(YAC_Object *) const;

   //set Vector length to Scale
   /* @method unitScale,float s
      Set length to s
      @arg s New length of vector
    */
   YM void _unitScale_YAC_RSELF(sF32);
   YM void _unitScale_YAC_RVAL(sF32, YAC_Value *);
   YM void _unitScale_YAC_RARG(sF32, YAC_Object *) const;

   //init element values
   /* @method init,float a,float b
      Initialise values to x=a and y=b
      @arg a new x value
      @arg b new y value
    */
   YM void _init(sF32, sF32);
   /* @method initScalef,Vector2f v,float s
      Initialise values to s*v
      @arg v Vector2f instance
      @arg s scale to multiply
    */
   YM void _initScalef(YAC_Object *, sF32);
   /* @method New,float a,float b:Vector2f
      Initialise and return new Vector2f instance with x=a and y=b
      @arg a new x value
      @arg b new y value
      @return new Vector2f instance
    */
   YM void _New(sF32,sF32,YAC_Value*);
};

/* @function vector2f,float a,float b:Vector2f
   Initialise and return new Vector2f instance with x=a and y=b
   @arg a new x value
   @arg b new y value
   @return new Vector2f instance
 */
YF void YAC_CALL _vector2f(sF32,sF32,YAC_Value*);

#endif
