/// vector3f.cpp
///
/// (c) 2008-2023 by Carsten Busse <carsten.busse@googlemail.com>,
///                  Bastian Spiegel <bs@tkscript.de> (additional coding)
///     - Distributed under terms of the Lesser GNU General Public License (LGPL).
///       See COPYING and <http://www.gnu.org/licenses/licenses.html#LGPL> for further information.
///
///

#include <math.h>

#include "tkmath_yac.h"
#include "tkmath.h"
#include "matrix.h"
#include "vector2f.h"
#include "vector3f.h"
#include "matrix3f.h"

#include "ying_Math.h"

// _Vector3f& _Vector3f::operator=(const _Vector3f& obj) {
//    floats[0] = obj.floats[0];
//    floats[1] = obj.floats[1];
//    floats[2] = obj.floats[2];
//    return *this;
// }

_Vector3f::_Vector3f(void) {
   floats[0] = floats[1] = floats[2] = 0.0f;
}

_Vector3f::~_Vector3f() {
}

sUI _Vector3f::yacArrayGetNumElements(void) {
   return 3;
}

sUI _Vector3f::yacArrayGetMaxElements(void) {
   return 3;
}

sUI _Vector3f::yacArrayGetElementType(void) {
   return YAC_TYPE_FLOAT;
}

sUI _Vector3f::yacArrayGetElementByteSize(void) {
   return sizeof(sF32);
}

void _Vector3f::yacArraySet(YAC_ContextHandle _ctx, sUI _idx, YAC_Value *_val) {
   if(_idx < 3) 
   {
      sF32 v;
      switch (_val->type)
      {
         case YAC_TYPE_VOID:
            v = 0.0f;
            break;
         case YAC_TYPE_INT:
            v = (sF32) _val->value.int_val;
            break;
         case YAC_TYPE_FLOAT:
            v = _val->value.float_val;
            break;
         case YAC_TYPE_OBJECT:
            if(YAC_VALID(_val->value.object_val)) 
            {
               if(!_val->value.object_val->yacScanF32(&v))
               {
                  v = 0.0f;
               }
            } 
            else 
            {
               // Silently cast <null> to 0.0f
               v = 0.0f;
            }
            break;
         default:
            // Should never ever happen!
            Dyac_throw(TypeMismatch, "tkmath::Vector3f::yacArraySet: unknown type");
            v = 0.0f;
            break;
      }
      floats[_idx] = v;
   } 
   else 
   {
      static char buf[256];
      Dyac_snprintf(buf, 256, "tkmath::Vector3f::yacArraySet Index too high (%d>2)", _idx);
      Dyac_throw(WriteArrayOutOfBounds, buf);
   }
}

void YAC_VCALL _Vector3f::yacArrayGet(YAC_ContextHandle _ctx, sUI _idx, YAC_Value *_val) {
   if(_idx < 3) 
   {
      _val->initFloat(floats[_idx]);
   } 
   else 
   {
      static char buf[256];
      Dyac_snprintf(buf, 256, "tkmath::Vector3f::yacArrayGet: Index too high (%d>2)", _idx);
      _val->initFloat(0.0f);
      Dyac_throw(ReadArrayOutOfBounds, buf);
   }
}

void *YAC_VCALL _Vector3f::yacArrayGetPointer(void) {
   return (void*)floats;
}

sSI YAC_VCALL _Vector3f::yacTensorRank(void) {
   return YAC_TENSOR_RANK_VECTOR;
}

void YAC_VCALL _Vector3f::yacSerialize(YAC_Object *_ofs, sUI _rtti) {
   YAC_BEG_SERIALIZE();
   YAC_SERIALIZE_F32(floats[0]);
   YAC_SERIALIZE_F32(floats[1]);
   YAC_SERIALIZE_F32(floats[2]);
}

sUI YAC_VCALL _Vector3f::yacDeserialize(YAC_Object *_ifs, sUI _rtti) {
   YAC_BEG_DESERIALIZE();
   floats[0]=YAC_DESERIALIZE_F32();
   floats[1]=YAC_DESERIALIZE_F32();
   floats[2]=YAC_DESERIALIZE_F32();
   return 1;
}

void YAC_VCALL _Vector3f::yacOperatorClamp(YAC_Object *_min, YAC_Object *_max) {
   if(YAC_CHK(_min, clid_Vector3f))
   {
      if(YAC_CHK(_max, clid_Vector3f))
      {
         _Vector3f *min=(_Vector3f*)_min;
         _Vector3f *max=(_Vector3f*)_max;
         if (floats[0]<min->floats[0]) floats[0]=min->floats[0];
         else if (floats[0]>max->floats[0]) floats[0]=max->floats[0];
         if (floats[1]<min->floats[1]) floats[1]=min->floats[1];
         else if (floats[1]>max->floats[1]) floats[1]=max->floats[1];
         if (floats[2]<min->floats[2]) floats[2]=min->floats[2];
         else if (floats[2]>max->floats[2]) floats[2]=max->floats[2];
      }
      else 
      {
         Dyac_throw_def(InvalidPointer, "tkmath::Vector3f::yacOperatorClamp: invalid _max object.");
      }
   } 
   else 
   {
      Dyac_throw_def(InvalidPointer, "tkmath::Vector3f::yacOperatorClamp: invalid _min object.");
   }
}

void YAC_VCALL _Vector3f::yacOperatorWrap(YAC_Object *_min, YAC_Object *_max) {
   if(YAC_CHK(_min, clid_Vector3f))
   {
      if(YAC_CHK(_max, clid_Vector3f))
      {
         _Vector3f *min=(_Vector3f*)_min;
         _Vector3f *max=(_Vector3f*)_max;
         if (floats[0]<min->floats[0]) floats[0]+=(max->floats[0]-min->floats[0]);
         else if (floats[0]>max->floats[0]) floats[0]-=(max->floats[0]-min->floats[0]);
         if (floats[1]<min->floats[1]) floats[1]+=(max->floats[1]-min->floats[1]);
         else if (floats[1]>max->floats[1]) floats[1]-=(max->floats[1]-min->floats[1]);
         if (floats[2]<min->floats[2]) floats[2]+=(max->floats[2]-min->floats[2]);
         else if (floats[2]>max->floats[2]) floats[2]-=(max->floats[2]-min->floats[2]);
      }
      else 
      {
         Dyac_throw_def(InvalidPointer, "tkmath::Vector3f::yacOperatorWrap: invalid _max object.");
      }
   } 
   else 
   {
      Dyac_throw_def(InvalidPointer, "tkmath::Vector3f::yacOperatorWrap: invalid _min object.");
   }
}

void YAC_VCALL _Vector3f::yacOperator(sSI _cmd, YAC_Object *_o, YAC_Value *_r) {
   if(YAC_CHK(_o, clid_Vector3f)) 
   {
      _Vector3f *o=(_Vector3f*)_o;
      switch(_cmd) 
      {
         case YAC_OP_CEQ:
            YAC_RETI( Dfltequal(floats[0], o->floats[0]) && 
                      Dfltequal(floats[1], o->floats[1]) && 
                      Dfltequal(floats[2], o->floats[2]) );
            break;
         case YAC_OP_CNE:
            YAC_RETI( Dfltnotequal(floats[0], o->floats[0]) || 
                      Dfltnotequal(floats[1], o->floats[1]) || 
                      Dfltnotequal(floats[2], o->floats[2]) );
            break;
         case YAC_OP_ASSIGN:
            floats[0] = o->floats[0];
            floats[1] = o->floats[1];
            floats[2] = o->floats[2];
            break;
         case YAC_OP_ADD:
            _add_YAC_RSELF(_o);
            break;
         case YAC_OP_SUB:
            _sub_YAC_RSELF(_o);
            break;
         case YAC_OP_MUL:
            _mul_YAC_RSELF(_o);
            break;
         case YAC_OP_NEG:
            floats[0] = -floats[0];
            floats[1] = -floats[1];
            floats[2] = -floats[2];
            break;
         case YAC_OP_INIT:
            floats[0] = 0.0f;
            floats[1] = 0.0f;
            floats[2] = 0.0f;
            break;
         default:
         {
            // Should never ever happen !
            static char buf[256];
            Dyac_snprintf(buf, 256, "tkmath::Vector3f::yacOperator unsupported cmd %d", _cmd);
            Dyac_throw_def(TkMathUnsupportedOperator, buf);
         }
         return;
      } // switch _cmd
   } 
   else if(YAC_VALID(_o))
   {
      sF32 val;
      if (_o->yacScanF32(&val)) {
         yacOperatorF32(_cmd, val, _r);
      } else
      // Check whether the object is derived from FloatArray
      if (YAC_Is_FloatArray(_o)) 
      {
         YAC_FloatArray *o=(YAC_FloatArray*)_o;
         switch (_cmd) 
         {
            case YAC_OP_ASSIGN:
               assignFloatArray(o);
               break;

            default:
            {
               // Create temporary Vector3f object on stack
               _Vector3f t;
               t.class_ID = clid_Vector3f;
               t.assignFloatArray(o);

               // Recursively call yacOperator again
               yacOperator(_cmd, &t, _r);
            }
            break;
         }
      }
      else 
      {
         // Handle generic float container object
         switch(_cmd)
         {
            case YAC_OP_ASSIGN:
               assignGeneric(NULL, _o);
               break;

            default:
            {
/*               _Vector3f t;
               t.class_ID = clid_Vector3f;
               t.assignGeneric(NULL, _o);*/

               // Recursively call yacOperator again
               YAC_Object::yacOperator(_cmd, _o, _r);
            }
            break;
         }
      }
   }
   else
   {
      switch (_cmd) {
      case YAC_OP_NEG:
         floats[0]=-floats[0];
         floats[1]=-floats[1];
         floats[2]=-floats[2];
	 break;
      default:
      // Call generic yacOperator, e.g. for <null> comparisons
            YAC_Object::yacOperator(_cmd, _o, _r);
      }
   }

}

void _Vector3f::yacOperatorI(sSI _cmd, sSI val, YAC_Value *_r) {
   yacOperatorF32(_cmd, (sF32)val, _r);
}

void _Vector3f::yacOperatorF32(sSI _cmd, sF32 val, YAC_Value *_r) {
   switch(_cmd)
   {
      case YAC_OP_ASSIGN:
         init(val, 0.0f, 0.0f);
         break;

      case YAC_OP_ADD:
         _addf_YAC_RSELF(val);
         break;

      case YAC_OP_SUB:
         _addf_YAC_RSELF(-val);
         break;

      case YAC_OP_MUL:
         _mulf_YAC_RSELF(val);
         break;

      case YAC_OP_DIV:
         if(Dfltnonzero(val))
         {
            _mulf_YAC_RSELF(1.0f/val);
         }
         else
         {
            Dyac_throw_def(TkMathDivisionByZero, "tkmath::Vector3f::yacOperatorF32 OP_DIV Cannot divide by zero");
         }
         break;

      default:
         char buf[256];
         Dyac_snprintf(buf, 256, "tkmath::Vector3f::yacOperatorF32 Operator not supported %d", _cmd);
         Dyac_throw_def(TkMathUnsupportedOperator,buf);
         break;
   }
}

void _Vector3f::assignFloats(sF32 *_fv, sUI _ne) {
   if(_ne <= 3)
   {
      // Copy as much elements from array as possible but never more than 3
      sUI i;
      for(i=0; i<_ne; i++)
      {
         floats[i] = _fv[i];
      }
      // Fill the rest with 0.0f
      for(; i<3; i++)
      {
         floats[i] = 0.0f;
      }
   }
   else
   {
      // Copy three elements
      floats[0] = _fv[0];
      floats[1] = _fv[1];
      floats[2] = _fv[2];
   }
}

void _Vector3f::assignFloatArray(YAC_FloatArray *_fa) {
   assignFloats(_fa->elements, _fa->num_elements);
}

void _Vector3f::assignGeneric(YAC_ContextHandle _ctx, YAC_Object *_o) {
   sUI ne = _o->yacArrayGetNumElements();
   if(ne > 0)
   {
      if(YAC_TYPE_FLOAT == _o->yacArrayGetElementType())
      {
         sF32 *fv = (sF32*) _o->yacArrayGetPointer();
         if(NULL != fv)
         {
            // Copy fast
            assignFloats(fv, ne);
            return;
         }
         else
         {
            // Retrieve individual float elements and assign
            YAC_Value v;
            sUI i;
            for(i=0; (i<3) && (i<ne); i++)
            {
               _o->yacArrayGet(_ctx, i, &v);
               floats[i] = v.value.float_val;
            }
            // Fill the rest with 0.0f
            for(; i<3; i++)
            {
               floats[i] = 0.0f;
            }
         }
      }
      else
      {
         // Retrieve individual elements, typecast to float and assign
         YAC_Value v;
         sUI i;
         for(i=0; (i<3) && (i<ne); i++)
         {
            _o->yacArrayGet(_ctx, i, &v);
            v.typecast(YAC_TYPE_FLOAT);
            floats[i] = v.value.float_val;
            v.unset();
         }
         // Fill the rest with 0.0f
         for(; i<3; i++)
         {
            floats[i] = 0.0f;
         }
      }
   }
   else
   {
      // Other array is empty or no array. Silently assign 0.0f
      floats[0] = floats[1] = floats[2] = 0.0f;
   }
}

void _Vector3f::init(sF32 _x, sF32 _y, sF32 _z) {
   floats[0] = _x;
   floats[1] = _y;
   floats[2] = _z;
}

void _Vector3f::_setX(sF32 x) {
   floats[0] = x;
}

void _Vector3f::_setY(sF32 y) {
   floats[1] = y;
}

void _Vector3f::_setZ(sF32 z) {
   floats[2] = z;
}

sF32 _Vector3f::_getX(void) const {
   return floats[0];
}

sF32 _Vector3f::_getY(void) const {
   return floats[1];
}

sF32 _Vector3f::_getZ(void) const {
   return floats[2];
}

void _Vector3f::_getXy_YAC_RSELF(void) const {
}

void _Vector3f::_getXy_YAC_RVAL(YAC_Value *_r) const {
   _Vector2f *o = YAC_NEW_POOLED(Vector2f);
   o->init(floats[0], floats[1]);
   _r->initObject(o, 1);
}

void _Vector3f::_getXy_YAC_RARG(YAC_Object *_o) const {
   if(YAC_CHK(_o, clid_Vector2f))
   {
      YAC_CAST_ARG(_Vector2f, o, _o);
      o->init(floats[0], floats[1]);
   }
}

sBool YAC_VCALL _Vector3f::yacToString(YAC_String *s) const {
#ifdef YAC_BIGSTRING
   s->yacArrayAlloc(1024,0,0,0);
   s->printf("(%g, %g, %g)",floats[0],floats[1],floats[2]);
   s->fixLength();
   return 1;
#endif
   return 0;
}

void _Vector3f::_getString(YAC_Value *_r) const {
   YAC_String *s=YAC_New_String();
   yacToString(s);
   YAC_RETS(s,1);
}

void _Vector3f::_add_YAC_RSELF(YAC_Object *_o) {
   if (YAC_CHK(_o,clid_Vector3f)) {
      _Vector3f *o=(_Vector3f*)_o;
      //add all elements
      floats[0]+=o->floats[0];
      floats[1]+=o->floats[1];
      floats[2]+=o->floats[2];
   } else {
/*      // Add generic Object
      YAC_Value r;
      yacOperator(YAC_OP_ADD, _o, &r);*/
      Dyac_throw_def(NativeClassTypeMismatch,"tkmath::Vector3f::add_SELF Object not of Type Vector3f");
   }
}

void _Vector3f::_add_YAC_RVAL(YAC_Object *_o, YAC_Value *_r) {
   if (YAC_CHK(_o,clid_Vector3f)) {
      _Vector3f *o=(_Vector3f*)_o;
      _Vector3f *r=YAC_NEW_POOLED(Vector3f);
      //add all elements
      r->floats[0] = floats[0] + o->floats[0];
      r->floats[1] = floats[1] + o->floats[1];
      r->floats[2] = floats[2] + o->floats[2];
      _r->initObject(r,1);
   } else {
/*      // Add generic Object
      YAC_Value r;
      yacOperator(YAC_OP_ADD, _o, &r);*/
      Dyac_throw_def(NativeClassTypeMismatch,"tkmath::Vector3f::add_VAL Object not of Type Vector3f");
   }
}

void _Vector3f::_add_YAC_RARG(YAC_Object *_o, YAC_Object *_r) const {
   if (YAC_CHK(_o,clid_Vector3f)&&YAC_CHK(_r,clid_Vector3f)) {
      _Vector3f *o=(_Vector3f*)_o;
      _Vector3f *r=(_Vector3f*)_r;
      //add all elements
      r->floats[0] = floats[0] + o->floats[0];
      r->floats[1] = floats[1] + o->floats[1];
      r->floats[2] = floats[2] + o->floats[2];
   } else {
/*      // Add generic Object
      YAC_Value r;
      yacOperator(YAC_OP_ADD, _o, &r);*/
      Dyac_throw_def(NativeClassTypeMismatch,"tkmath::Vector3f::add_ARG Object not of Type Vector3f");
   }
}

void _Vector3f::_sub_YAC_RSELF(YAC_Object *_o) {
   if (YAC_CHK(_o,clid_Vector3f)) {
      _Vector3f *o=(_Vector3f*)_o;
      //sub all elements
      floats[0]-=o->floats[0];
      floats[1]-=o->floats[1];
      floats[2]-=o->floats[2];
   } else {
/*      // Subtract generic Object
      YAC_Value r;
      yacOperator(YAC_OP_ADD, _o, &r);*/
      Dyac_throw_def(NativeClassTypeMismatch,"tkmath::Vector3f::sub_SELF Object not of Type Vector3f");
   }
}

void _Vector3f::_sub_YAC_RVAL(YAC_Object *_o, YAC_Value *_r) {
   if (YAC_CHK(_o,clid_Vector3f)) {
      _Vector3f *o=(_Vector3f*)_o;
      _Vector3f *r=YAC_NEW_POOLED(Vector3f);
      //sub all elements
      r->floats[0] = floats[0] - o->floats[0];
      r->floats[1] = floats[1] - o->floats[1];
      r->floats[2] = floats[2] - o->floats[2];
      _r->initObject(r,1);
   } else {
/*      // Subtract generic Object
      YAC_Value r;
      yacOperator(YAC_OP_ADD, _o, &r);*/
      Dyac_throw_def(NativeClassTypeMismatch,"tkmath::Vector3f::sub_VAL Object not of Type Vector3f");
   }
}

void _Vector3f::_sub_YAC_RARG(YAC_Object *_o, YAC_Object *_r) const {
   if (YAC_CHK(_o,clid_Vector3f)&&YAC_CHK(_r,clid_Vector3f)) {
      _Vector3f *o=(_Vector3f*)_o;
      _Vector3f *r=(_Vector3f*)_r;
      //sub all elements
      r->floats[0] = floats[0] - o->floats[0];
      r->floats[1] = floats[1] - o->floats[1];
      r->floats[2] = floats[2] - o->floats[2];
   } else {
/*      // Subtract generic Object
      YAC_Value r;
      yacOperator(YAC_OP_ADD, _o, &r);*/
      Dyac_throw_def(NativeClassTypeMismatch,"tkmath::Vector3f::sub_ARG Object not of Type Vector3f");
   }
}

sF32 _Vector3f::_dot(YAC_Object *_o) {
   if (YAC_CHK(_o,clid_Vector3f)) {
      _Vector3f *o=(_Vector3f*)_o;
      //returns scalar dot product
      return (floats[0]*o->floats[0]+floats[1]*o->floats[1]+floats[2]*o->floats[2]);
   } else {
      Dyac_throw_def(NativeClassTypeMismatch, "tkmath::Vector3f::dot Object not of Type Vector3f");
   }
   return 0.0f;
}

void _Vector3f::_cross_YAC_RVAL(YAC_Object *_o, YAC_Value *_r) {
   if (YAC_CHK(_o,clid_Vector3f)) {
      _Vector3f *o=(_Vector3f*)_o;
      _Vector3f *res = YAC_NEW_POOLED(Vector3f);
      //returns the orthogonal vector of the cross product
      res->floats[0]=floats[1]*o->floats[2]-floats[2]*o->floats[1];
      res->floats[1]=floats[2]*o->floats[0]-floats[0]*o->floats[2];
      res->floats[2]=floats[0]*o->floats[1]-floats[1]*o->floats[0];
      _r->initObject(res,1);
   } else {
      Dyac_throw_def(NativeClassTypeMismatch, "tkmath::Vector3f::cross Object not of type Vector3f");
   }
}

void _Vector3f::_cross_YAC_RARG(YAC_Object *_o, YAC_Object *_r) const {
   if (YAC_CHK(_o,clid_Vector3f)&&YAC_CHK(_r,clid_Vector3f)) {
      _Vector3f *o=(_Vector3f*)_o;
      _Vector3f *res = (_Vector3f*)_r;
      //returns the orthogonal vector of the cross product
      res->floats[0]=floats[1]*o->floats[2]-floats[2]*o->floats[1];
      res->floats[1]=floats[2]*o->floats[0]-floats[0]*o->floats[2];
      res->floats[2]=floats[0]*o->floats[1]-floats[1]*o->floats[0];
   } else {
      Dyac_throw_def(NativeClassTypeMismatch, "tkmath::Vector3f::cross Object not of type Vector3f");
   }
}

sF32 _Vector3f::_triple(YAC_Object *_b, YAC_Object *_c) {
   if (YAC_CHK(_b,clid_Vector3f)&&YAC_CHK(_c,clid_Vector3f)) {
      _Vector3f *b=(_Vector3f*)_b;
      _Vector3f *c=(_Vector3f*)_c;
      _Matrix3f m;
      //return the scalar triple product (a*(bxc))
      memcpy(m.floats+0/*(0*sizeof(sF32))*/,floats,sizeof(sF32)*3); //we copy as row vector instead of col
      memcpy(m.floats+3/*(3*sizeof(sF32))*/,b->floats,sizeof(sF32)*3); //because the determinant of a
      memcpy(m.floats+6/*(6*sizeof(sF32))*/,c->floats,sizeof(sF32)*3); //transposed matrix is the same
      return m._det();
   } else {
      Dyac_throw_def(NativeClassTypeMismatch, "tkmath::Vector3f::triple Object not of type Vector3f");
   }
   return 0.0f;
}

void _Vector3f::_tensor_YAC_RVAL(YAC_Object *_o, YAC_Value *_r) {
   if(YAC_CHK(_o, clid_Vector3f))
   {
      _Vector3f t;
      _Matrix3f *m = YAC_NEW_POOLED(Matrix3f);
      _Vector3f *o=(_Vector3f*)_o;
      //return the matrix of the tensor product

      m->floats[0] = floats[0] * o->floats[0];
      m->floats[1] = floats[0] * o->floats[1];
      m->floats[2] = floats[0] * o->floats[2];
      m->floats[3] = floats[1] * o->floats[0];
      m->floats[4] = floats[1] * o->floats[1];
      m->floats[5] = floats[1] * o->floats[2];
      m->floats[6] = floats[2] * o->floats[0];
      m->floats[7] = floats[2] * o->floats[1];
      m->floats[8] = floats[2] * o->floats[2];
      _r->initObject(m,1);
   }
   else 
   {
      Dyac_throw_def(NativeClassTypeMismatch, "tkmath::Vector3f::tensor: invalid Vector3f object");
   }
}

void _Vector3f::_tensor_YAC_RARG(YAC_Object *_o, YAC_Object *_m) const {
   if(YAC_CHK(_m, clid_Matrix3f))
   {
      _Vector3f t;
      _Matrix3f *m = (_Matrix3f*) _m;
      _Vector3f *o;
      if(YAC_CHK(_o, clid_Vector3f)) 
      {
         o = (_Vector3f*) _o;
         //return the matrix of the tensor product
         m->floats[0] = floats[0] * o->floats[0];
         m->floats[1] = floats[0] * o->floats[1];
         m->floats[2] = floats[0] * o->floats[2];
         m->floats[3] = floats[1] * o->floats[0];
         m->floats[4] = floats[1] * o->floats[1];
         m->floats[5] = floats[1] * o->floats[2];
         m->floats[6] = floats[2] * o->floats[0];
         m->floats[7] = floats[2] * o->floats[1];
         m->floats[8] = floats[2] * o->floats[2];
      }
      else
      {
/*         t.assignGeneric(NULL, _o);
           o = &t;*/
         Dyac_throw_def(NativeClassTypeMismatch,"tkmath::Vector3f::tensor_: invalid Vector3f Object");
      }
   }
   else 
   {
      Dyac_throw_def(NativeClassTypeMismatch, "tkmath::Vector3f::tensor_: invalid Matrix3f object");
   }
}

void _Vector3f::_addf_YAC_RSELF(sF32 s) {
   floats[0] += s;
   floats[1] += s;
   floats[2] += s;
}

void _Vector3f::_addf_YAC_RVAL(sF32 s, YAC_Value *_r) {
   _Vector3f *r = YAC_NEW_POOLED(Vector3f);
   r->floats[0] = floats[0] + s;
   r->floats[1] = floats[1] + s;
   r->floats[2] = floats[2] + s;
   _r->initObject(r, 1);
}

void _Vector3f::_addf_YAC_RARG(sF32 s, YAC_Object *_r) const {
   if(YAC_CHK(_r, clid_Vector3f))
   {
      YAC_CAST_ARG(_Vector3f, r, _r);
      r->floats[0] = floats[0] + s;
      r->floats[1] = floats[1] + s;
      r->floats[2] = floats[2] + s;
   }
   else
   {
      Dyac_throw_def(NativeClassTypeMismatch, "tkmath::Vector3f::addf_ARG Return object not of type Vector3f");
   }
}

void _Vector3f::_mulf_YAC_RSELF(sF32 s) {
   // multiply elements by scalar value
   floats[0] *= s;
   floats[1] *= s;
   floats[2] *= s;
}

void _Vector3f::_mulf_YAC_RVAL(sF32 s, YAC_Value *_r) {
   // multiply elements by scalar value
   _Vector3f *r=YAC_NEW_POOLED(Vector3f);
   r->floats[0] = floats[0] * s;
   r->floats[1] = floats[1] * s;
   r->floats[2] = floats[2] * s;
   _r->initObject(r,1);
}

void _Vector3f::_mulf_YAC_RARG(sF32 s, YAC_Object *_r) const {
   // multiply elements by scalar value
   if (YAC_CHK(_r, clid_Vector3f)) {
      _Vector3f *r=(_Vector3f*)_r;
      r->floats[0] = floats[0] * s;
      r->floats[1] = floats[1] * s;
      r->floats[2] = floats[2] * s;
   } else {
      Dyac_throw_def(NativeClassTypeMismatch, "tkmath::Vector3f::mulf_ARG Return type not of type Vector3f");
   }
}

void _Vector3f::_mul_YAC_RSELF(YAC_Object *_o) {
   if(YAC_CHK(_o, clid_Vector3f)) 
   {
      // multiply each element
      _Vector3f *o=(_Vector3f*)_o;
      floats[0] *= o->floats[0];
      floats[1] *= o->floats[1];
      floats[2] *= o->floats[2];
   } 
   else 
   {
/*      // Multiply by generic Object
      YAC_Value r;
      yacOperator(YAC_OP_MUL, _o, &r);*/
      Dyac_throw_def(NativeClassTypeMismatch,"tkmath::Vector3f::mul_SELF Invalid Vector3f object");

   }
}

void _Vector3f::_mul_YAC_RVAL(YAC_Object *_o, YAC_Value *_r) {
   if(YAC_CHK(_o, clid_Vector3f)) 
   {
   //multiply each element
      _Vector3f *o=(_Vector3f*)_o;
      _Vector3f *r=YAC_NEW_POOLED(Vector3f);
      r->floats[0] = floats[0] * o->floats[0];
      r->floats[1] = floats[1] * o->floats[1];
      r->floats[2] = floats[2] * o->floats[2];
      _r->initObject(r,1);
   } 
   else 
   {
/*      // Multiply by generic Object
      YAC_Value r;
      yacOperator(YAC_OP_MUL, _o, &r);*/
      Dyac_throw_def(NativeClassTypeMismatch,"tkmath::Vector3f::mul_VAL Invalid Vector3f object");

   }
}

void _Vector3f::_mul_YAC_RARG(YAC_Object *_o, YAC_Object *_r) const {
   if(YAC_CHK(_o, clid_Vector3f)&&YAC_CHK(_r, clid_Vector3f)) 
   {
   //multiply each element
      _Vector3f *o=(_Vector3f*)_o;
      _Vector3f *r=(_Vector3f*)_r;
      r->floats[0] = floats[0] * o->floats[0];
      r->floats[1] = floats[1] * o->floats[1];
      r->floats[2] = floats[2] * o->floats[2];
   } 
   else 
   {
/*      // Multiply by generic Object
      YAC_Value r;
      yacOperator(YAC_OP_MUL, _o, &r);*/
      Dyac_throw_def(NativeClassTypeMismatch,"tkmath::Vector3f::mul_ARG Invalid Vector3f object");

   }
}

sF32 _Vector3f::_getAbs(void) const {
   return sqrtf(floats[0]*floats[0] + floats[1]*floats[1] + floats[2]*floats[2]);
}

sF32 _Vector3f::_getAbsSqr(void) const {
   return (floats[0]*floats[0] + floats[1]*floats[1] + floats[2]*floats[2]);
}

void _Vector3f::_unit_YAC_RSELF(void) {
   sF32 absval = _getAbs();
   if(Dfltnonzero(absval)) 
   {
      absval=1.0f/absval;
      floats[0] *= absval; 
      floats[1] *= absval;
      floats[2] *= absval;
   } 
   else 
   {
      //Dyac_throw_def(TkMathDivisionByZero, "tkmath::Vector3f::unit_SELF Zero Vector cannot be normalized");
   }
}

void _Vector3f::_unit_YAC_RVAL(YAC_Value *_r) {
   sF32 absval = _getAbs();
   if(Dfltnonzero(absval)) 
   {
      _Vector3f *r=YAC_NEW_POOLED(Vector3f);
      absval=1.0f/absval;
      r->floats[0] = floats[0] * absval; 
      r->floats[1] = floats[1] * absval;
      r->floats[2] = floats[2] * absval;
      _r->initObject(r,1);
   } 
   else 
   {
      //Dyac_throw_def(TkMathDivisionByZero, "tkmath::Vector3f::unit_VAL Zero Vector cannot be normalized");
   }
}

void _Vector3f::_unit_YAC_RARG(YAC_Object *_r) const {
   if (YAC_CHK(_r, clid_Vector3f)) {
      sF32 absval = _getAbs();
      if(Dfltnonzero(absval)) 
      {
         _Vector3f *r = (_Vector3f*)_r;
         absval=1.0f/absval;
         r->floats[0] = floats[0] * absval; 
         r->floats[1] = floats[1] * absval;
         r->floats[2] = floats[2] * absval;
      } 
      else 
      {
         //Dyac_throw_def(TkMathDivisionByZero, "tkmath::Vector3f::unit_ARG Zero Vector cannot be normalized");
      }
   } else {
     Dyac_throw_def(NativeClassTypeMismatch, "tkmath::Vector3f::unit_ARG Return object not of type Vector3f");
   }
}

void _Vector3f::_unitScale_YAC_RSELF(sF32 s) {
   sF32 absval = _getAbs();
   if(Dfltnonzero(absval)) 
   {
      absval = s / absval;

      floats[0] *= absval;
      floats[1] *= absval;
      floats[2] *= absval;
   }
   else 
   {
      //Dyac_throw_def(TkMathDivisionByZero, "tkmath::Vector3f::unitScale_SELF Zero Vector cannot be normalized");
   }
}

void _Vector3f::_unitScale_YAC_RVAL(sF32 s, YAC_Value *_r) {
   sF32 absval = _getAbs();
   if(Dfltnonzero(absval)) 
   {
      _Vector3f *r=YAC_NEW_POOLED(Vector3f);
      absval = s / absval;

      r->floats[0] = floats[0] * absval;
      r->floats[1] = floats[1] * absval;
      r->floats[2] = floats[2] * absval;
      _r->initObject(r,1);
   }
   else 
   {
      //Dyac_throw_def(TkMathDivisionByZero, "tkmath::Vector3f::unitScale_VAL Zero Vector cannot be normalized");
   }
}

void _Vector3f::_unitScale_YAC_RARG(sF32 s, YAC_Object *_r) const {
   if (YAC_CHK(_r,clid_Vector3f)) {
      sF32 absval = _getAbs();
      if(Dfltnonzero(absval)) 
      {
         _Vector3f *r=(_Vector3f*)_r;
         absval = s / absval;

         r->floats[0] = floats[0] * absval;
         r->floats[1] = floats[1] * absval;
         r->floats[2] = floats[2] * absval;
      }
      else 
      {
         //Dyac_throw_def(TkMathDivisionByZero, "tkmath::Vector3f::unitScale_ARG Zero Vector cannot be normalized");
      }
   } else {
      Dyac_throw_def(NativeClassTypeMismatch, "tkmath::Vector3f::unitScale_ARG Return object not of type Vector3f");
   }
}

//initialising methods

void _Vector3f::_init(sF32 x, sF32 y, sF32 z) {
   floats[0] = x;
   floats[1] = y;
   floats[2] = z;
}

void _Vector3f::_initScalef(YAC_Object *_o, sF32 _s) {
   if(YAC_CHK(_o, clid_Vector3f))
   {
      _Vector3f *o = (_Vector3f*) _o;

      floats[0] = o->floats[0] * _s;
      floats[1] = o->floats[1] * _s;
      floats[2] = o->floats[2] * _s;
   }
   else
   {
      Dyac_throw_def(InvalidPointer, "tkmath::Vector3f::initScalef: invalid object parameter");
   }
}

sF32 _Vector3f::_distanceToPlane(YAC_Object *_q, YAC_Object *_n) const {
   if(YAC_CHK(_q, clid_Vector3f))
   {
      if(YAC_CHK(_n, clid_Vector3f))
      {
         YAC_CAST_ARG(_Vector3f, q, _q);
         YAC_CAST_ARG(_Vector3f, n, _n);
         YAC_NEW_STACK(Vector3f, v);
         _sub_YAC_RARG(q, &v);
         return (v._dot(n) / n->_getAbs());
      }
   }
	return 0.0f;
}

void _Vector3f::_New(sF32 x, sF32 y, sF32 z, YAC_Value *_r) {
   _Vector3f *r=YAC_NEW_POOLED(Vector3f);
   r->_init(x,y,z);
   _r->initObject(r,1);
}

void YAC_CALL _vector3f(sF32 x, sF32 y, sF32 z,YAC_Value *_r) {
   _Vector3f *r=YAC_NEW_POOLED(Vector3f);
   r->_init(x,y,z);
   _r->initObject(r,1);
}
