/// vector4f.h
///
/// (c) 2008-2020 by Carsten Busse <carsten.busse@googlemail.com>,
///                  Bastian Spiegel <bs@tkscript.de> (additional coding)
///     - Distributed under terms of the Lesser GNU General Public License (LGPL).
///       See COPYING and <http://www.gnu.org/licenses/licenses.html#LGPL> for further information.
///
///

#ifndef __VECTOR4F_H_
#define __VECTOR4F_H_


YG("Math");

/* @class Vector4f,Object
This class represents a single precision 4 dimensional euclidean space vector.
 */
YC class _Vector4f : public YAC_Object {
  public:
   sF32 floats[4];

   public:
   _Vector4f(void);
   ~_Vector4f();

   YAC_POOLED(_Vector4f,YAC_POOL_PRIORITY_HIGH);

   sUI   YAC_VCALL yacArrayGetNumElements     (void);
   sUI   YAC_VCALL yacArrayGetMaxElements     (void);
   sUI   YAC_VCALL yacArrayGetElementType     (void);
   sUI   YAC_VCALL yacArrayGetElementByteSize (void);
   void  YAC_VCALL yacArraySet                (YAC_ContextHandle, sUI, YAC_Value *);
   void  YAC_VCALL yacArrayGet                (YAC_ContextHandle, sUI, YAC_Value *);
   void *YAC_VCALL yacArrayGetPointer         (void);
   sSI   YAC_VCALL yacTensorRank              (void);
   void  YAC_VCALL yacSerialize               (YAC_Object *, sUI);
   sUI   YAC_VCALL yacDeserialize             (YAC_Object *, sUI);
   void  YAC_VCALL yacOperatorClamp           (YAC_Object *, YAC_Object *);
   void  YAC_VCALL yacOperatorWrap            (YAC_Object *, YAC_Object *);
   //supports:
   //YAC_OP_CEQ, YAC_OP_CNE, YAC_OP_ASSIGN, YAC_OP_ADD = _add,
   //YAC_OP_SUB = _sub, YAC_OP_MUL = _mul, YAC_OP_NEG, YAC_OP_INIT
   void  YAC_VCALL yacOperator                (sSI, YAC_Object *, YAC_Value *);
   void  YAC_VCALL yacOperatorI               (sSI, sSI, YAC_Value *);
   void  YAC_VCALL yacOperatorF32             (sSI, sF32, YAC_Value *);
   void  YAC_VCALL yacOperatorAssign          (YAC_Object *_robj);
   sUI   YAC_VCALL yacOperatorPriority        (void) { return YAC_OP_PRIO_FLOAT; }
   sBool YAC_VCALL yacToString                (YAC_String*) const;

  public:
   // Helper methods
   void assignFloats(sF32 *_fv, sUI _ne);
   void assignFloatArray(YAC_FloatArray *_fa);
   void assignGeneric(YAC_ContextHandle _ctx, YAC_Object *_o);

  public:

   /* @method init,float x,float y,float z,float w
      Initialize vector
    */
   YM void init(sF32 _x, sF32 _y, sF32 _z, sF32 _w);

   /* @method setX,float a
      Set x element to a
      @arg a new value for x
    */
   YM void _setX(sF32);


   /* @method setY,float a
      Set y element to a
      @arg a new value for y
    */
   YM void _setY(sF32);


   /* @method setZ,float a
      Set z element to a
      @arg a new value for z
    */
   YM void _setZ(sF32);


   /* @method setW,float a
      Set w element to a
      @arg a new value for w
    */
   YM void _setW(sF32);


   /* @method getX:float
      Get value of x element
      @return value of x element
    */
   YM sF32 _getX(void);


   /* @method getY:float
      Get value of y element
      @return value of y element
    */
   YM sF32 _getY(void);


   /* @method getZ:float
      Get value of z element
      @return value of z element
    */
   YM sF32 _getZ(void);


   /* @method getW:float
      Get value of w element
      @return value of w element
    */
   YM sF32 _getW(void);

   /* @method getXYZ:float
      Get xyz vector
      @return xyz vector
    */
   YM void _getXyz_YAC_RSELF(void) const;
   YM void _getXyz_YAC_RVAL(YAC_Value *_r) const;
   YM void _getXyz_YAC_RARG(YAC_Object *_o) const;

   //produce string representation
   /* @method getString:String
      Get string representation of vector (x, y, z, w)
      @return string representation of vector
    */
   YM void _getString(YAC_Value *) const;

   //arithmetic
   /* @method add,Vector4f v
      Add another vector v to this instance
      @arg v vector to add
    */
   YM void _add_YAC_RSELF(YAC_Object *);
   YM void _add_YAC_RVAL(YAC_Object *, YAC_Value *);
   YM void _add_YAC_RARG(YAC_Object *, YAC_Object *) const;

   /* @method sub,Vector4f v
      Substract another vector v from this instance
      @arg v vector to substract
    */
   YM void _sub_YAC_RSELF(YAC_Object *);
   YM void _sub_YAC_RVAL(YAC_Object *, YAC_Value *);
   YM void _sub_YAC_RARG(YAC_Object *, YAC_Object *) const;

   //scalar dot product
   /* @method dot,Vector4f v:float
      Calculate dot product from this instance and v
      @arg v Vector4f instance
      @return scalar value
    */
   YM sF32 _dot(YAC_Object *);

   //tensor product, produces a Matrix4f out of two Vector4f. Result is written to _retM4f
   /* @method tensor,Vector4f v:Matrix4f
      Create resulting matrix of tensor product from this instance and v
      @arg a Vector4f instance
      @return Matrix4f instance
    */
   YM void _tensor_YAC_RVAL(YAC_Object *, YAC_Value *);
   YM void _tensor_YAC_RARG(YAC_Object *_v4f, YAC_Object *_retM4f) const;

   /* @method addf,float v
      Add scalar value to all components
      @arg v scalar value
    */
   YM void _addf_YAC_RSELF(sF32);
   YM void _addf_YAC_RVAL(sF32, YAC_Value *);
   YM void _addf_YAC_RARG(sF32, YAC_Object *) const;

   /* @method mulf,float v
      Multiply this instance by scalar value v
      @arg v scalar value
    */
   YM void _mulf_YAC_RSELF(sF32);
   YM void _mulf_YAC_RVAL(sF32, YAC_Value *);
   YM void _mulf_YAC_RARG(sF32, YAC_Object *) const;
   
   //multiplies each element individual (not really mathematics, but sometimes useful)
   /* @method mul,Vector4f v
      Multiply the x element of this instance by the x element of v and do the same with the y, z and w elements
      @arg v Vector4f instance
    */
   YM void _mul_YAC_RSELF(YAC_Object *);
   YM void _mul_YAC_RVAL(YAC_Object *, YAC_Value *);
   YM void _mul_YAC_RARG(YAC_Object *, YAC_Object *) const;

   /* @method getAbs:float
      Return length of vector
      @return length of vector
    */
   YM sF32 _getAbs(void) const;

   /* @method getAbsSqr:float
      Return squared length of vector
      @return length<sup>2</sup> of vector
    */
   YM sF32 _getAbsSqr(void) const;

   //scale vector to length 1 if possible
   /* @method unit
      Set length to 1
    */
   YM void _unit_YAC_RSELF(void);
   YM void _unit_YAC_RVAL(YAC_Value *);
   YM void _unit_YAC_RARG(YAC_Object *) const;

   //set Vector length to Scale
   /* @method unitScale,float s
      Set length to s
      @arg s new length of vector
    */
   YM void _unitScale_YAC_RSELF(sF32);
   YM void _unitScale_YAC_RVAL(sF32, YAC_Value *);
   YM void _unitScale_YAC_RARG(sF32, YAC_Object *) const;

   //init element values
   /* @method init,float a,float b,float c,float d
      Initialise values to x=a, y=b, z=c and w=d
      @arg a new x value
      @arg b new y value
      @arg c new z value
      @arg d new w value
    */
   YM void _init(sF32, sF32, sF32, sF32);


   /* @method initScalef,Vector4f v,float s
      Initialise vector to s*v
      @arg v Vector4f instance
      @arg s scale to multiply
    */
   YM void _initScalef(YAC_Object *_o, sF32 _s);


   /* @method rotate,Matrix4f m
      Multiply matrix by this vector and replace this vector by the result

      This is a convenience method to remove the need for temporary Vector4f variables in scripts (the temporary is created on the stack instead).

      @arg m Matrix4f instance (rotation matrix)
    */
   YM void _rotate(YAC_Object *_m);


   /* calc distance of vertex to plane defined by Vector4f q (on plane) and plane normal n */
   YM sF32 _distanceToPlane (YAC_Object *_q, YAC_Object *_n) const;

   /* @method New,float a,float b,float c,float d:Vector4f
      Initialise and return new Vector4f instance with x=a, y=b, z=c and w=d
      @arg a new x value
      @arg b new y value
      @arg c new z value
      @arg d new w value
    */
   YM void _New(sF32, sF32, sF32, sF32, YAC_Value *);
};

/* @function vector4f,float a,float b,float c,float d:Vector4f
   Initialise and return new Vector4f instance with x=a, y=b, z=c and w=d
   @arg a new x value
   @arg b new y value
   @arg c new z value
   @arg d new w value
 */
YF void YAC_CALL _vector4f(sF32, sF32, sF32, sF32,YAC_Value*);

#endif  // __VECTOR4F_H_
